#!/bin/bash
# grapher for networks from Maude terms

#
# Copyright (c) 2011, SRI International .
# Developed with the sponsorship of the Defense Advanced Research Projects Agency DARPA .
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this data, including any software or models in source or binary form, as well as any drawings, specifications, and documentation collectively "the Data", to deal in the Data without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Data, and to permit persons to whom the Data is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Data .
#
# THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR OTHER DEALINGS IN THE DATA .
#

## requires Graphviz 2.x (tested with 2.28 on Mac OS X):
##   http://www.graphviz.org/Download_macos.php
##
## use in interactive mode (default):
## - Copy and paste a term from Maude.  This can contain the period at the end but no other
##   lines or the eq definition.  You can grab the output from a command 
##     Maude> red net5 .
##   command to display that graph (so you don't have to go into the source code).
## - Once you have copied the text into your terminal, hit CTRL-D to tell it to parse the input.
##   It should pop up a Preview window with the graph.  
##   The script loops, so it is ready for more input afterwards.
## - If you want to quit, you can input empty text or, at any time, use CTRL-C to stop the script.
##
## use in batch mode (-b):
##   $> cat net.txt | ./graph.sh -b | open -f -a /Applications/Preview.app 
## or
##   $> cat net.txt | ./graph.sh -b > net.png
## and open file net.png in your viewer 

################################################################################
# functions:

usage () {
    if [ $# -gt 0 ]; then
	echo " *** ERROR: $1"
    fi
    cat << EOF

usage: $0 [<OPTIONS>] 

Generate graph from Maude term of a NetworkConfiguration.  User is prompted to 
paste term into terminal.

OPTIONS:
   -h              Show this message
   -c <COLOR>      Use given color for dataflow nodes (default is "red").  If 
                   "none" is given, skip dataflow nodes.  See 
                     http://www.graphviz.org/doc/info/colors.html
                     http://www.graphviz.org/doc/info/attrs.html#k:color
                   for details on color names and specification
   -b              Run once in batch mode assuming term comes from STDIN
   -t <DOT-OUTPUT> Use given string as argument for dot's -T parameter (default
                   is "png")
EOF
}

generate () {
    # remove any trailing period
    # turn components and df into lines without leading newlines (for awk)
    # remove empty lines
    # sort so that dataflows will be parsed after components
    # break up dataflows:
    #  1) df name in first line
    #  2) replace "conn ( N |" with "conn(N) |" in dataflow
    #  3) each connection on new line ", conn(" => "\nconn(" 
    # remove failure probablity operator " pr(F)" => " F" 
    # remove spaces in front of colons
    # replace multiple spaces with one
    # remove any spaces inside brackets: "( bla  )" => "(bla)"
    # build graph structure with awk 
    sed 's/\.[ 	]*$//' | sed 's/[ 	]*\(<[^<]*>\)[ 	]*/\
\1\
/g' | sed 's/[ 	]*df[ 	]*(/\
df(/g' | sed '/^[ 	]*$/d' | sort | sed 's/^df([ 	]*\([0-9][0-9]*\)[ 	]*\|[ 	]*\([a-z][a-z]*\)[ 	]*\|/< df(\1) - \2\
/' | sed '/^[^<]/ s/conn[ 	]*(\([ 	]*[0-9][0-9]*\)/conn(\1)/g' | sed 's/,[ 	]*conn(/\
 conn(/g' | sed 's/[ 	]*pr(\([^)]*\))/ \1/g' | sed 's/[ 	]*:[ 	]*/: /g' | sed 's/[ 	][ 	]*/ /g' | sed 's/([ 	]*\([^ 	)][^ 	)]*\)[ 	]*)/(\1)/g' | awk -f parse.awk -v color=$COLOR
}


################################################################################
# parse command line:

COLOR="red"
DOTOUT="png"
BATCH=""
while getopts "hc:bt:" OPTION; do
    case $OPTION in
        h|\?)
            usage; exit 1
            ;;
	c)
	    COLOR=$OPTARG
	    ;;
	b)
	    BATCH=1
	    ;;
	t)
	    DOTOUT=$OPTARG
	    ;;
    esac
    shift $((OPTIND-1)); OPTIND=1 
done


################################################################################
# if batch mode, just run once over STDIN where newlines are turned into spaces

if [ "$BATCH" ]; then
    tr '\n' ' ' | generate | dot -T$DOTOUT 
    exit 0
fi


################################################################################
# loop infinitely reading user input until input is empty 

while true; do  

    # obtain term from user:
    echo "Copy and paste term here (finish term with ^D on new line; exit script with empty input):"
    TERM=$(cat)

    if [ -z "$TERM" ]; then  # if input is empty, exit
	echo "Goodbye."
	exit 0
    fi

    echo $TERM | generate | dot -T$DOTOUT | open -f -a /Applications/Preview.app

done

exit 0

# --- code for testing:
generate <<EOF
< conn(0) | Start to PH.in0 > < conn(1) | PH.out1 to PL(1).in0 > < conn(2) | PL(1).out1 to PE.in0 > < conn(3) | PE.out1 to PL(3).in0 > < conn(5) | PL(3).out1 to SW(0).in0 > < conn(7) | SW(0).out1 to PL(7).in0 > < conn(16) | PL(7).out1 to BE(3).in0 > < conn(17) | BE(3).out1 to PL(17).in0 > < conn(18) | PL(17).out1 to BH(3).in0 > < conn(19) | BH(3).out1 to End > df(1 | tt |   conn(0 |    om : pr(1.0), com : pr(1.0), vSA : pr(1.0), vDA : pr(1.0), vSN : pr(1.0), vLen : pr(1.0), vData : pr(1.0),    vFCS : pr(1.0), te : pr(1.0), tl : pr(1.0) ),   conn(1 |    om : pr(1.0e-3), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0), vData : pr(    0.0), vFCS : pr(0.0), te : pr(0.0), tl : pr(0.0) ),   conn(2 |    om : pr(1.0e-3), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0), vData : pr(    1.0000000000000009e-3), vFCS : pr(0.0), te : pr(0.0), tl : pr(0.0) ),   conn(3 |    om : pr(2.8972009000000076e-3), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0),    vData : pr(1.000000000000001e-4), vFCS : pr(1.0000000000000001e-1), te : pr(0.0), tl : pr(0.0) ),   conn(5 |    om : pr(2.8972009000000076e-3), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0),    vData : pr(1.0999000000000425e-3), vFCS : pr(1.0000000000000001e-1), te : pr(0.0), tl : pr(0.0) ),   conn(7 |    om : pr(1.0438743739069634e-1), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0),    vData : pr(0.0), vFCS : pr(0.0), te : pr(0.0), tl : pr(0.0) ),   conn(16 |    om : pr(1.0438743739069634e-1), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0),    vData : pr(1.0000000000000009e-3), vFCS : pr(0.0), te : pr(0.0), tl : pr(0.0) ),   conn(17 |    om : pr(1.0608829520834773e-1), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0),    vData : pr(1.000000000000001e-4), vFCS : pr(1.0000000000000001e-1), te : pr(0.0), tl : pr(0.0) ),   conn(18 |    om : pr(1.0608829520834773e-1), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0),    vData : pr(1.0999000000000425e-3), vFCS : pr(1.0000000000000001e-1), te : pr(0.0), tl : pr(0.0) ),   conn(19 |    om : pr(1.0e-3), com : pr(0.0), vSA : pr(0.0), vDA : pr(0.0), vSN : pr(0.0), vLen : pr(0.0), vData : pr(    0.0), vFCS : pr(0.0), te : pr(0.0), tl : pr(0.0) ) ) < pl(1) | PL(1).in0 | PL(1).out1 | plfail : pr(1.0e-3) > < pl(3) | PL(3).in0 | PL(3).out1 | plfail : pr(1.0e-3) > < pl(7) | PL(7).in0 | PL(7).out1 | plfail : pr(1.0e-3) > < pl(17) | PL(17).in0 | PL(17).out1 | plfail : pr(1.0e-3) > < hi-sw(0) | SW(0).in0, SW(0).in3, SW(0).in4, SW(0).in5, SW(0).in6 | SW(0).out0, SW(0).out1, SW(  0  ).out2, SW(    0).out3, SW(0 ).out7 | swfail : pr(1.0e-3), fcschfail : pr(1.0000000000000001e-1) > < hi-es(1) | PE.in0 | PE.out1 | esfail : pr(1.0e-3), fcschfail : pr(1.0000000000000001e-1) > < hi-es(53) | BE(3).in0, BE(3).in2, BE(3).in3 | BE(3).out1, BE(3).out3, BE(3).out4 | esfail : pr(1.0e-3),    fcschfail : pr(1.0000000000000001e-1) > < hi-host(1) | PH.in0 | PH.out1 | hostfail : pr(1.0e-3) > < hi-host(53) | BH(3).in0, BH(3).in1, BH(3).in2 | BH(3).out0, BH(3).out1, BH(3).out2 | hostfail : pr(1.0e-3) >
EOF
exit 0
