#!/usr/bin/awk -f
# parse Maude term with components on each line into dot notation
# expects variable "color" passed from command line

#
# Copyright (c) 2011, SRI International .
# Developed with the sponsorship of the Defense Advanced Research Projects Agency DARPA .
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this data, including any software or models in source or binary form, as well as any drawings, specifications, and documentation collectively "the Data", to deal in the Data without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Data, and to permit persons to whom the Data is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Data .
#
# THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR OTHER DEALINGS IN THE DATA .
#

BEGIN { 
    FS = "[\|<>]"
    DF_NODE = ""  # current dataflow node

    printf "digraph net {\n\tgraph [rankdir=LR];\n\tnode [shape=record,fontname=\"Helvetica\",fontsize=8,fixedsize=false];\n\tedge [fontname=\"Helvetica\",fontsize=8,weight=100.0];\n\n"
}

# trim given string from leading and trailing white space; if string then is "empty" return ""
function trim(s) {
    gsub(/^[ \t]+|[ \t]+$/,"",s)
    if (s ~ /empty/)
	return ""
    return s
}

# remove any parenthesis from port
function reduce_port(port) {
    gsub(/\(|\)|\./,"",port)
    return port
}

# convert set of ports into dot label
function port_set(set,ports) {
    n = split(set,arr,"[ \t]*,[ \t]*")
    portstring = ""
    for (i = 1; i <= n; i++) {
	port = reduce_port(arr[i])
	portstring = portstring "<" port "> " arr[i] " |"
	ports[port] = component  # remember component name of port
    }
    return substr(portstring,1,length(portstring)-1)  # remove last "|" character
}

# round failure value to 1 digit after decimal point for numeric values (unless it's zero)
function pretty_print_failures(failureset) {
    n = split(failureset,arr,"(: )|(, )")
    values = ""
    if (n == 1)
	values = failureset
    else
	for (i = 1; i < n ; i += 2)
	    if (arr[i+1] ~ /(true)|(false)/)
		values = values sprintf("%s: %s\\l",arr[i],arr[i+1])
	    else
		if (arr[i+1] == 0)
		    values = values sprintf("%s: 0\\l",arr[i])
		else
		    values = values sprintf("%s: %.1e\\l",arr[i],arr[i+1])
    return values
}

{
    if ($1 ~ /^[ \t]*conn\([0-9]+\)/) {  # connection of dataflow
	if (DF_NODE != "" && color != "none") {
	    CN = trim($1)
	    NODE = sprintf("\"%s-%s\"",CN,DF_NODE)
	    DFs[NODE] = CN  # collect dataflow nodes and their connection
	    FAILURES = trim($2)
	    sub("[)]+[ \t]*[\.]*$","",FAILURES)  # remove any trailing brackets and periods
	    printf "\t%s [fontcolor=%s,color=%s,label=\"{ { %s | %s } }\"];\n",NODE,color,color,DF_NODE,pretty_print_failures(FAILURES)
	}
    } else 
	if ($2 ~ /conn/)  # connection declaration
	    CNs[trim($2)] = trim($3)  # collect connections in associative array
	else 
	    if ($2 ~ /df\([0-9]+\) - (be|tt|rc)/)   # start new dataflow node
		DF_NODE = trim($2)
	    else {  # parse a component
		component = trim($2)
		# print node line:
		printf "\t\"%s\" [label=\"{ {%s} | { %s | %s } | {%s} }\"];\n",component,port_set(trim($3),IPs),component,pretty_print_failures(trim($5)),port_set(trim($4),OPs)
	    }
}

# evaluate connection end: does is denote component or do we need to add a point node as end?
function conn_end(port,ports) {
#    printf("// conn end: %s \n",port)
    if (port in ports)
	node = "\"" ports[port] "\":" port
    else {
	node = port
	if (!(node in points)) {
	    points[node] = 1
	    printf "\t%s [shape=point];\n",node
	}
    }
    return node
}

END {
    # print edges (and potential end points if component doesn't exist):
    print ""
    for (conn in CNs) {
	split(CNs[conn], arr, "[ \t]+to[ \t]+")
	start = conn_end(reduce_port(arr[1]),OPs)
	end = conn_end(reduce_port(arr[2]),IPs)
	printf "\t%s -> %s [label=\"%s\"];\n",start,end,conn
	# remember start and end of each connection for dataflow edges:
	STARTs[conn] = start
	ENDs[conn] = end
    }
    # print invisible edges between dataflow nodes:
    printf "\n\tedge [weight=0.0];\n"
    for (df in DFs) {
	printf "\t%s:e -> %s:nw [dir=none,style=invis];\n",STARTs[DFs[df]],df
	printf "\t%s:ne -> %s:w [dir=none,style=invis];\n",df,ENDs[DFs[df]]
    }
    # finish graph:
    print "}"
}
