/*
 * Copyright (c) 2011, SRI International .
 * Developed with the sponsorship of the Defense Advanced Research Projects Agency DARPA .
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this data, including any software or models in source or binary form, as well as any drawings, specifications, and documentation collectively "the Data", to deal in the Data without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Data, and to permit persons to whom the Data is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Data .
 *
 * THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR OTHER DEALINGS IN THE DATA .
 */
package com.sri.promise.gui;

import com.sri.promise.gui.cmdline.Grapher;
import com.sri.promise.gui.cmdline.MaudeRunner;
import com.sri.promise.gui.cmdline.SchedulerRunner;
import com.sri.promise.logging.LevelOptionHandler;
import com.sri.promise.logging.LogConfiguration;
import org.kohsuke.args4j.Argument;
import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.filechooser.FileFilter;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;
import java.util.prefs.Preferences;

/**
 * @author linda
 */
public final class GUI extends JFrame implements DocumentListener {

    // static initializations
    private final Preferences preferences = Preferences.userNodeForPackage(this.getClass());
    private final static String KEY_LAST_DIR = "last directory";
    private final static String KEY_LAST_DIVIDER_V = "last V divider location";
    private final static String KEY_LAST_DIVIDER_H = "last H divider location";
    private final static String KEY_LAST_WIDTH = "last width of window";
    private final static String KEY_LAST_HEIGHT = "last height of window";
    private final static String KEY_LAST_X = "last X of window";
    private final static String KEY_LAST_Y = "last Y of window";
    private final static String KEY_LAST_VLIDS = "last number of VL IDs";
    static final Logger LOGGER = Logger.getLogger(GUI.class.getName());
    private final static String SHOW_ACTION_NAME = "show graph";
    private final static String EVAL_TEXT = "red evalConn(    ,";

    // Maude lib and exe paths (to be obtained from command line)
    private final String maudeLib;
    private final String maudeExe;
    private final String schedulerDir;

    // UI components
    private final JFileChooser fileChooser = new JFileChooser();
    private final JTextField fileField = new JTextField();
    private final JTextField termField = new JTextField();
    private final JTextField connField = new JTextField();
    private final JLabel evalLabel = new JLabel(EVAL_TEXT);
    private final JButton runNetButton = new JButton(new AbstractAction("Run") {
        private static final long serialVersionUID = -2689637880373945871L;
        public void actionPerformed(ActionEvent event) {
            evalLabel.setText("red evalConn("+termField.getText().trim()+",");
            analysisPane.outPrintln("Running maude reducing term \"" + termField.getText() + "\"...");
            new MaudeRunner(maudeExe, maudeLib, fileField.getText(), termField.getText(), textPane, analysisPane).execute();
        }
    });
    private final JButton runConnButton = new JButton(new AbstractAction("Run") {
        private static final long serialVersionUID = -2689637880373945871L;
        public void actionPerformed(ActionEvent event) {
            analysisPane.outPrintln("Running maude to analyze connection \"" + connField.getText() + "\"...");
            new MaudeRunner(maudeExe, maudeLib, fileField.getText(), "evalConn("+termField.getText().trim()+","+connField.getText()+")", null, analysisPane).execute();
        }
    });
    private final InputTextArea textPane = new InputTextArea();
    private final GraphArea graphArea = new GraphArea();
    private final AnalysisArea analysisPane = new AnalysisArea();
    private final Action showAction = new AbstractAction("Show") {
        private static final long serialVersionUID = -2689637880373945871L;
        public void actionPerformed(ActionEvent actionEvent) {
            analysisPane.outPrintln("Generating graph from text...");
            // obtain selected text or (if nothing selected) all text
            String text = textPane.getSelectedText();
            if (text == null)
                text = textPane.getText();
            // send to grapher, which will update the graph area when done
            new Grapher(text, graphArea, analysisPane).execute();
        }
    };
    // UI component initialization
    {
        fileChooser.setFileFilter(new FileFilter() {
            @Override
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true; // accept directories to browse
                return file.getName().endsWith(".maude");
            }
            @Override
            public String getDescription() {
                return "Maude files";
            }
        });
        fileChooser.setCurrentDirectory(new File(Preferences.userNodeForPackage(this.getClass()).get(
                KEY_LAST_DIR, System.getProperty("user.dir"))));
        runNetButton.setEnabled(false); // gets enabled when both text fields "file" and "net" contain text
        runConnButton.setEnabled(false); // gets enabled when all three text fields "file", "net" and "conn" contain text
        // listen for document changes on text fields to enable buttons and set current directory
        fileField.getDocument().addDocumentListener(this);
        termField.getDocument().addDocumentListener(this);
        connField.getDocument().addDocumentListener(this);
        // run maude when hitting ENTER in term field:
        termField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                GUI.this.runNetButton.doClick();
            }
        });
        // run maude when hitting ENTER in conn field:
        connField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                GUI.this.runConnButton.doClick();
            }
        });
        // ENTER in text area causes graph update
        textPane.getActionMap().put(SHOW_ACTION_NAME, showAction);
        textPane.getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, InputEvent.SHIFT_DOWN_MASK), SHOW_ACTION_NAME);
    }

    public GUI(String maudeLib, String maudeExe, String schedulerDir, File file) throws HeadlessException {
        super("PROMISE GUI");
        this.maudeLib = maudeLib;
        this.maudeExe = maudeExe;
        this.schedulerDir = schedulerDir;

        // process optional file argument
        if (file != null)
            try {
                fileField.setText(file.getCanonicalPath());
            } catch (IOException e) {
                LOGGER.severe("Cannot get canonical path of given file "+file.getAbsolutePath()+" due to "+e.getMessage());
            }

        // put everything together
        JPanel contentPane = createPanes();
        contentPane.setOpaque(true); //content panes must be opaque
        setContentPane(contentPane);

        addWindowListener(new WindowAdapter() {
            @Override
            public void windowOpened(WindowEvent windowEvent) {
                GUI.this.setBounds(
                        preferences.getInt(KEY_LAST_X, 0),
                        preferences.getInt(KEY_LAST_Y, 0),
                        preferences.getInt(KEY_LAST_WIDTH, 1000),
                        preferences.getInt(KEY_LAST_HEIGHT, 650));
                LOGGER.config("Window opened: " + GUI.this.getSize() + " at " + GUI.this.getLocation());
                // after this resizing and moving events can now be recorded:
                GUI.this.addComponentListener(new ComponentAdapter() {
                    @Override
                    public void componentMoved(ComponentEvent componentEvent) {
                        preferences.putInt(KEY_LAST_X, GUI.this.getX());
                        preferences.putInt(KEY_LAST_Y, GUI.this.getY());
                        LOGGER.config("Window position: " + GUI.this.getLocation());
                    }

                    @Override
                    public void componentResized(ComponentEvent componentEvent) {
                        preferences.putInt(KEY_LAST_WIDTH, GUI.this.getWidth());
                        preferences.putInt(KEY_LAST_HEIGHT, GUI.this.getHeight());
                        LOGGER.config("Window size: " + GUI.this.getSize());
                    }
                });
            }

            @Override
            public void windowClosing(WindowEvent windowEvent) {
                // save size and location
                preferences.putInt(KEY_LAST_WIDTH, GUI.this.getWidth());
                preferences.putInt(KEY_LAST_HEIGHT, GUI.this.getHeight());
                preferences.putInt(KEY_LAST_X, GUI.this.getX());
                preferences.putInt(KEY_LAST_Y, GUI.this.getY());
                LOGGER.config("Window closing: " + GUI.this.getSize() + " at " + GUI.this.getLocation());
            }
        });
    }

    private JLabel makeMonoLabel(String text, Font font) {
        JLabel label = new JLabel(text);
        label.setFont(new Font("Monospaced", font.getStyle(), font.getSize()));
        return label;
    }

    private JPanel createPanes() {
        // -- Maude file and term selection pane
        JPanel maudePane = new JPanel(new GridBagLayout());
        Font font = maudePane.getFont();
        GridBagConstraints c = new GridBagConstraints();
        c.anchor = GridBagConstraints.LINE_START;
        c.gridx = 0;
        c.gridy = 0;
        maudePane.add(new JLabel("Maude file: "), c);
        c.gridy = 1;
        maudePane.add(new JLabel("Maude term: "), c);
        c.gridy = 2;
        maudePane.add(new JLabel("Maude term: "), c);

        c.gridx = 2;
        c.gridy = 0;
        maudePane.add(new JButton(new AbstractAction("Choose...") {
            private static final long serialVersionUID = -7527199490504455906L;
            public void actionPerformed(ActionEvent event) {
                if (fileChooser.showOpenDialog(GUI.this) == JFileChooser.APPROVE_OPTION) {
                    File file = fileChooser.getSelectedFile();
                    String path = file.getAbsolutePath();
                    fileField.setText(path);
                    analysisPane.outPrintln("Maude file "+path+" chosen.");
                }
            }
        }), c);
        c.gridy = 1;
        maudePane.add(runNetButton, c);
        c.gridy = 2;
        maudePane.add(runConnButton, c);

        c.fill = GridBagConstraints.HORIZONTAL;
        c.weightx = 1.0; // text inputs get all the space
        c.gridx = 1;
        c.gridy = 0;
        fileField.setFont(new Font("Monospaced", font.getStyle(), font.getSize()));
        maudePane.add(fileField, c);
        c.gridy = 1;
        JPanel termPane = new JPanel(new BorderLayout(2, 0));
        termPane.add(makeMonoLabel("red", font), BorderLayout.LINE_START);
        termField.setFont(new Font("Monospaced", font.getStyle(), font.getSize()));
        termPane.add(termField, BorderLayout.CENTER);
        termPane.add(makeMonoLabel(".", font), BorderLayout.LINE_END);
        maudePane.add(termPane, c);
        c.gridy = 2;
        JPanel connPane = new JPanel(new BorderLayout(2, 0));
        evalLabel.setFont(new Font("Monospaced", font.getStyle(), font.getSize()));
        connPane.add(evalLabel, BorderLayout.LINE_START);
        connField.setFont(new Font("Monospaced", font.getStyle(), font.getSize()));
        connPane.add(connField, BorderLayout.CENTER);
        connPane.add(makeMonoLabel(") .", font), BorderLayout.LINE_END);
        maudePane.add(connPane, c);

        // -- text input area in scroll pane
        JScrollPane textScrollPane = new JScrollPane(textPane);
        textScrollPane.setPreferredSize(new Dimension(800, 300));

        // -- graph output in scroll pane; slider below
        JPanel graphOutput = new JPanel(new BorderLayout(0, 5));
        JScrollPane graphScrollPane = new JScrollPane(graphArea);
        graphScrollPane.setPreferredSize(new Dimension(800, 300));
        graphScrollPane.setBackground(null);
        graphScrollPane.getViewport().setBackground(null);
        graphOutput.add(graphScrollPane, BorderLayout.CENTER);
        graphOutput.add(graphArea.getSlider(), BorderLayout.PAGE_END);

        // -- split pane between text area and graph
        final JSplitPane splitPaneV = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
                textScrollPane, graphOutput);
        splitPaneV.setDividerLocation(preferences.getInt(KEY_LAST_DIVIDER_V, 0));
        splitPaneV.addPropertyChangeListener(JSplitPane.DIVIDER_LOCATION_PROPERTY,
                new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent e) {
                        preferences.putInt(KEY_LAST_DIVIDER_V, splitPaneV.getDividerLocation());
                        LOGGER.config("Vertical divider location: "+splitPaneV.getDividerLocation());
                    }});
        splitPaneV.setBorder(null);

        // -- TT scheduler and RC checker
        JPanel ttPane = new JPanel();
        ttPane.setLayout(new BoxLayout(ttPane, BoxLayout.LINE_AXIS));
        ttPane.add(new JLabel("schedule/check:"));

        JRadioButton radioButton;
        final ButtonGroup buttonGroup = new ButtonGroup();
        ttPane.add(radioButton = new JRadioButton("tt"));
        radioButton.setSelected(true);
        radioButton.setActionCommand("tt");
        buttonGroup.add(radioButton);
        ttPane.add(radioButton = new JRadioButton("rc"));
        radioButton.setActionCommand("rc");
        buttonGroup.add(radioButton);

        final JSpinner spinner = new JSpinner(new SpinnerNumberModel(
                preferences.getInt(KEY_LAST_VLIDS, 5), 5, 100, 1));
        spinner.addChangeListener(new ChangeListener() {
            @Override
            public void stateChanged(ChangeEvent changeEvent) {
                preferences.putInt(KEY_LAST_VLIDS, (Integer) spinner.getValue());
            }
        });
        ttPane.add(spinner);
        ttPane.add(Box.createRigidArea(new Dimension(3, 0)));
        ttPane.add(new JLabel("VL IDs"));
        ttPane.add(new JButton(new AbstractAction("Run") {
            private static final long serialVersionUID = -2815860763379219582L;
            @Override
            public void actionPerformed(ActionEvent event) {
                analysisPane.outPrintln("Schedule/check "+buttonGroup.getSelection().getActionCommand()+
                        " with "+spinner.getValue()+" VL IDs...");
                new SchedulerRunner(schedulerDir, analysisPane,
                        buttonGroup.getSelection().getActionCommand().trim(),
                        (Integer) spinner.getValue()).execute();
            }
        }));
        ttPane.add(Box.createHorizontalGlue());

        // -- main panel
        JPanel mainPane = new JPanel(new BorderLayout());
        mainPane.add(maudePane, BorderLayout.PAGE_START);
        mainPane.add(splitPaneV, BorderLayout.CENTER);
        mainPane.add(ttPane, BorderLayout.PAGE_END);

        // -- analysis output in scroll pane
        JScrollPane analysisScrollPane = new JScrollPane(analysisPane);
        analysisScrollPane.setPreferredSize(new Dimension(300, 600));
        analysisScrollPane.getViewport().setBackground(null);
        analysisScrollPane.setBackground(null);

        // -- split pane between inputs and failure analysis output
        final JSplitPane splitPaneH = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                mainPane, analysisScrollPane);
        splitPaneH.setDividerLocation(preferences.getInt(KEY_LAST_DIVIDER_H, 0));
        splitPaneH.addPropertyChangeListener(JSplitPane.DIVIDER_LOCATION_PROPERTY,
                new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent e) {
                        preferences.putInt(KEY_LAST_DIVIDER_H, splitPaneH.getDividerLocation());
                        LOGGER.config("Divider location: "+splitPaneH.getDividerLocation());
                    }});
        splitPaneH.setBorder(null);

        // -- content pane
        JPanel contentPane = new JPanel(new BorderLayout(0, 5));
        contentPane.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
        contentPane.add(splitPaneH, BorderLayout.CENTER);
        return contentPane;
    }

    // true if both fields are not empty
    private boolean checkBothFields() {
        return !"".equals(fileField.getText()) && !"".equals(termField.getText());
    }

    // true if all three fields are not empty
    private boolean checkAllThreeFields() {
        return !"".equals(fileField.getText()) && !"".equals(termField.getText()) && !"".equals(connField.getText());
    }

    private void updateCurrentDir() {
        File file = new File(fileField.getText());
        if (file.exists()) {
            fileChooser.setCurrentDirectory(file.getParentFile());
            Preferences.userNodeForPackage(this.getClass()).put(KEY_LAST_DIR, file.getParent());
        }
    }

    // *** DocumentListener interface

    public void insertUpdate(DocumentEvent event) {
        runNetButton.setEnabled(checkBothFields());
        runConnButton.setEnabled(checkAllThreeFields());
        if (fileField.getDocument().equals(event.getDocument()))
            updateCurrentDir();
    }

    public void removeUpdate(DocumentEvent event) {
        runNetButton.setEnabled(checkBothFields());
        runConnButton.setEnabled(checkAllThreeFields());
        if (fileField.getDocument().equals(event.getDocument()))
            updateCurrentDir();
    }

    public void changedUpdate(DocumentEvent event) {
        // ignore
    }

    /**
     * Create the GUI and show it.  For thread safety,
     * this method should be invoked from the
     * event dispatch thread.
     *
     * @param frame Viewer instance to be displayed
     */
    private static void createAndShowGUI(GUI frame) {
        //Create and set up the window.
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        //Display the window.
        frame.pack();
        frame.setVisible(true);
    }

    private static void printUsage(PrintStream out, CmdLineParser parser) {
        out.println("usage: java -cp ... com.sri.promise.gui.GUI [options...] \nwith:");
        parser.printUsage(out);
    }

    public static void main(String[] args) {
        // parse arguments
        CmdLineParser.registerHandler(Level.class, LevelOptionHandler.class);
        final GUIOptions options = new GUIOptions();
        CmdLineParser parser = new CmdLineParser(options);
        try {
            parser.parseArgument(args);
        } catch (CmdLineException e) {
            System.err.println(e.getMessage());
            printUsage(System.err, parser);
            return;
        }

        if (options.displayHelp) {
            printUsage(System.out, parser);
            System.exit(1);
        }

        // configure logging
        try {
            LogConfiguration logConfig = new LogConfiguration();
            logConfig.setProperty("java.util.logging.ConsoleHandler.level",options.consoleLogLevel.getName());
            LogManager.getLogManager().readConfiguration(logConfig.asInputStream());
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Cannot configure logging", e);
        }

        final GUI frame = new GUI(options.mLib, options.mExe, options.sDir, options.file);

        //Schedule a job for the event dispatch thread:
        //creating and showing this application's GUI.
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                createAndShowGUI(frame);
            }
        });
    }

    static class GUIOptions {
        @Option(name = "-l", usage = "set console log level")
        Level consoleLogLevel = Level.CONFIG;

        @Option(name = "-h", usage = "display usage and exit")
        boolean displayHelp = false;

        @Option(name = "-mLib", required = true, usage = "path to MAUDE_LIB directory -- mandatory!", metaVar = "PATH")
        String mLib;

        @Option(name = "-mExe", usage = "path to Maude executable (default is $MAUDE_LIB/maude)", metaVar = "FILE")
        String mExe = mLib + File.separator+"maude";

        @Option(name = "-sDir", required = true, usage = "path to scheduler directory -- mandatory!", metaVar = "PATH")
        String sDir;

        @Argument(required = false, metaVar = "FILE", usage = "choose given Maude file")
        File file;
    }
}
