/*
 * Copyright (c) 2011, SRI International .
 * Developed with the sponsorship of the Defense Advanced Research Projects Agency DARPA .
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this data, including any software or models in source or binary form, as well as any drawings, specifications, and documentation collectively "the Data", to deal in the Data without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Data, and to permit persons to whom the Data is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Data .
 *
 * THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR OTHER DEALINGS IN THE DATA .
 */
package com.sri.promise.gui;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.MouseInputAdapter;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.util.Dictionary;
import java.util.Hashtable;

/**
 * @author linda
 */
public final class GraphArea extends JPanel {

    private BufferedImage image = null;
    private AffineTransform zoomTransform = new AffineTransform();
    private final JSlider slider = new JSlider(JSlider.HORIZONTAL, 1, 100, 100);
    private DragMover dragMover;

    public GraphArea() {
        slider.setEnabled(false);
        slider.setMajorTickSpacing(50);
        slider.setMinorTickSpacing(10);
        // create label table
        Dictionary<Integer,JLabel> labels = new Hashtable<Integer,JLabel>();
        labels.put(0, new JLabel("0%"));
        labels.put(50, new JLabel("50%"));
        labels.put(100, new JLabel("100%"));
        slider.setLabelTable(labels);
        slider.setPaintTicks(true);
        slider.setPaintLabels(true);
        slider.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent event) {
                JSlider source = (JSlider) event.getSource();
                if (!source.getValueIsAdjusting()) {
                    GraphArea.this.zoom(((double) source.getValue()) / 100.0d);
                }
            }
        });
        setBackground(null);
        setOpaque(true);
        addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent mouseEvent) {
                if (image != null && mouseEvent.getClickCount() > 1)
                    GraphArea.this.zoomToFit();
            }
        });
        setPreferredSize(new Dimension());
    }

    /**
     * Display graph from image given as argument.
     *
     * @param image Image to be displayed (could be NULL)
     */
    public void setGraph(BufferedImage image) {
        this.image = image;
        dragMover = null;
        setBackground(image == null?null:Color.WHITE);
        slider.setEnabled(image != null);
        if (image == null)
            slider.setValue(100);
        else {
            zoomToFit();
            dragMover = new DragMover(this);
        }
    }

    private void checkSize() {
        Dimension graphSize = new Dimension();
        // update size of graph area for scrolling
        if (image != null) {
            graphSize = new Dimension(
                    image.getWidth(),
                    image.getHeight());
            if (zoomTransform.getScaleX() != 1.0) {
                Shape s = zoomTransform.createTransformedShape(new Rectangle(graphSize));
                graphSize = s.getBounds().getSize();
            }
        }
        // update size:
        setPreferredSize(graphSize);
        revalidate();
    }

    private void zoom(double factor) {
        if (image != null) {
            zoomTransform = AffineTransform.getScaleInstance(factor, factor);
            if (getParent() instanceof JViewport) {
                // have to translate top-left point of view if embedded in JViewport
                JViewport viewport = (JViewport) getParent();
                Point2D location = zoomTransform.transform(viewport.getViewPosition(), null);
                viewport.setViewPosition(new Point(
                        (int) location.getX(),
                        (int) location.getY()));
            }
            checkSize();
            repaint();
        }
    }

    private void zoomToFit() {
        double zoomValue = Math.min(
                ((double) getParent().getWidth())/image.getWidth(),    // ratio of widths
                ((double) getParent().getHeight())/image.getHeight()); // ratio of heights
        if (zoomValue > 1.0d)
            zoomValue = 1.0d; // don't go past 100%
        if (zoomValue < 0.01d)
            zoomValue = 0.01d; // don't go below 1%
        slider.setValue((int) (100.0d * zoomValue));
    }

    public JSlider getSlider() {
        return slider;
    }

    @Override
    protected void paintComponent(Graphics graphics) {
        super.paintComponent(graphics);
        // paint graph, if image is set
        if (image != null) {
            Graphics2D g2d = (Graphics2D) graphics;
            g2d.drawImage(image, zoomTransform, null);
        }
    }
}
