/*
 * Copyright (c) 2011, SRI International .
 * Developed with the sponsorship of the Defense Advanced Research Projects Agency DARPA .
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this data, including any software or models in source or binary form, as well as any drawings, specifications, and documentation collectively "the Data", to deal in the Data without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Data, and to permit persons to whom the Data is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Data .
 *
 * THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR OTHER DEALINGS IN THE DATA .
 */
package com.sri.promise.gui.cmdline;

import com.sri.promise.gui.AnalysisArea;

import javax.swing.*;
import java.io.*;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Implements a Swing worker to run Maude from script with given values.
 *
 * @author linda
 */
public final class MaudeRunner extends SwingWorker<String, Void> {

    private static final Logger LOGGER = Logger.getLogger(MaudeRunner.class.getName());
    private static final String newline = System.getProperty("line.separator");
    private static final String SELECT_START = "result NetworkConfiguration:";
    private static final String SELECT_END = "Bye.";

    private final String maudeLib;
    private final String maudeExe;
    private final File maudeFile;
    private final String maudeTerm;
    private final JTextArea textArea;
    private final AnalysisArea outputArea;

    public MaudeRunner(String maudeExe, String maudeLib, String maudeFile, String maudeTerm, JTextArea textArea, AnalysisArea outputArea) {
        this.maudeExe = maudeExe;
        this.maudeLib = maudeLib;
        this.outputArea = outputArea;
        this.maudeFile = new File(maudeFile);
        this.maudeTerm = maudeTerm;
        this.textArea = textArea;
        // check inputs
        if (!new File(maudeLib).exists())
            throw new RuntimeException("Maude library does not exist");
        if (!new File(maudeExe).exists())
            throw new RuntimeException("Maude executable does not exist");
        if (!this.maudeFile.exists())
            throw new RuntimeException("Maude file does not exist");
        if (maudeTerm == null || "".equals(maudeTerm))
            throw new RuntimeException("Maude term to reduce is NULL or empty");
    }

    @Override
    protected String doInBackground() throws Exception {
        // obtain directory to execute in
        File directory = maudeFile.getParentFile();

        // start maude
        ProcessBuilder pb = new ProcessBuilder(maudeExe,
                "-interactive","-no-ansi-color","-no-wrap");
        pb.directory(directory);
        Map<String, String> env = pb.environment();
        env.put("MAUDE_LIB", maudeLib);
        final Process process = pb.start();

        // issue commands to maude:
        PrintWriter pw = new PrintWriter(process.getOutputStream());
        pw.write("load "+maudeFile.getCanonicalPath()+newline);
        pw.write("red "+maudeTerm+" ."+newline);
        pw.write("q"+newline);
        pw.close();

        // echo STDERR from process
        new Thread(new StdErrRedirect(outputArea, process.getErrorStream())).start();

        // start process to grab output of process:
        final StringBuilder output = new StringBuilder();
        Thread grabber = new Thread(new Runnable() {
            public void run() {
                BufferedReader reader = new BufferedReader(new InputStreamReader(process.getInputStream()));
                String line;
                try {
                    while ((line = reader.readLine()) != null)
                        if (textArea == null)
                            outputArea.outPrintln(line);
                        else
                            output.append(line+newline);
                } catch (IOException e) {
                    LOGGER.severe("While reading standard output from process: "+e.getMessage()); // stop
                }
            }
        });
        grabber.start();
        grabber.join();

        return output.toString();
    }

    @Override
    protected void done() {
        if (textArea != null)
            try {
                String text = get();
                textArea.setText(text);
                // select text between "result NetworkConfiguration:" and "Bye."
                int start = text.indexOf(SELECT_START);
                int end = text.lastIndexOf(SELECT_END);
                if (start >+ 0 && end >= 0) {
                    textArea.requestFocusInWindow(); // request focus to make selection visible
                    textArea.setSelectionStart(start + SELECT_START.length());
                    textArea.setSelectionEnd(end);
                }
            } catch (Exception ignore) {}
        // signal success
        outputArea.outPrintln("... success running Maude.");
    }
}
